"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var propertyNames_1 = require("./propertyNames");
var reducers_1 = require("./reducers");
var util_1 = require("./util");
var ITERATOR_SYMBOL = typeof Symbol !== "undefined" ? Symbol.iterator : "@@iterator";
/**
 * For compatibility with environments where common types aren't iterable.
 */
function getIterator(collection) {
    var anyCollection = collection;
    if (anyCollection[ITERATOR_SYMBOL]) {
        return anyCollection[ITERATOR_SYMBOL]();
    }
    else if (Array.isArray(anyCollection) ||
        typeof anyCollection === "string") {
        // Treat a string like an array of characters.
        return new ArrayIterator(anyCollection);
    }
    else {
        throw new Error("Cannot get iterator of non iterable value: " + anyCollection);
    }
}
exports.getIterator = getIterator;
/**
 * Iterator for arrays in environments without Iterable.
 */
var ArrayIterator = /** @class */ (function () {
    function ArrayIterator(array) {
        this.array = array;
        this.i = 0;
    }
    ArrayIterator.prototype[ITERATOR_SYMBOL] = function () {
        return this;
    };
    ArrayIterator.prototype.next = function () {
        var _a = this, i = _a.i, array = _a.array;
        if (i < array.length) {
            this.i++;
            return { done: false, value: array[i] };
        }
        else {
            return { done: true };
        }
    };
    return ArrayIterator;
}());
var RangeIterator = /** @class */ (function () {
    function RangeIterator(startOrEnd, end, step) {
        if (step === 0) {
            throw new Error("Step in rangeIterator() cannot be 0");
        }
        else if (end == null) {
            this.i = 0;
            this.end = startOrEnd;
            this.step = 1;
        }
        else if (step == null) {
            this.i = startOrEnd;
            this.end = end;
            this.step = 1;
        }
        else {
            this.i = startOrEnd;
            this.end = end;
            this.step = step;
        }
    }
    RangeIterator.prototype[ITERATOR_SYMBOL] = function () {
        return this;
    };
    RangeIterator.prototype.next = function () {
        var _a = this, i = _a.i, end = _a.end, step = _a.step;
        if ((step > 0 && i < end) || (step < 0 && i > end)) {
            var result = { done: false, value: i };
            this.i += step;
            return result;
        }
        else {
            return { done: true };
        }
    };
    return RangeIterator;
}());
function rangeIterator(startOrEnd, end, step) {
    // We can't satisfy the IterableIterator interface while functioning in
    // environments without Symbol, hence the cast.
    return new RangeIterator(startOrEnd, end, step);
}
exports.rangeIterator = rangeIterator;
/**
 * An iterable which enables lazy consumption of the output of a
 * transducer-based transform.
 */
var TransducerIterable = /** @class */ (function () {
    function TransducerIterable(iterator, xf) {
        this.iterator = iterator;
        this.upcoming = new ArrayIterator([]);
        this.hasSeenEnd = false;
        this.xfToArray = xf(reducers_1.toArray());
    }
    TransducerIterable.prototype[ITERATOR_SYMBOL] = function () {
        return this;
    };
    TransducerIterable.prototype.next = function () {
        while (true) {
            var backlogged = this.upcoming.next();
            if (!backlogged.done) {
                return backlogged;
            }
            else if (this.hasSeenEnd) {
                return { done: true };
            }
            else {
                var _a = this.iterator.next(), done = _a.done, value = _a.value;
                if (done) {
                    return { done: done };
                }
                else {
                    var outValues = this.xfToArray[propertyNames_1.STEP]([], value);
                    if (util_1.isReduced(outValues)) {
                        this.hasSeenEnd = true;
                        outValues = outValues[propertyNames_1.VALUE];
                    }
                    this.upcoming = new ArrayIterator(outValues);
                }
            }
        }
    };
    return TransducerIterable;
}());
function lazyTransduce(collection, transform) {
    // We can't satisfy the IterableIterator interface while functioning in
    // environments without Symbol, hence the cast.
    return new TransducerIterable(getIterator(collection), transform);
}
exports.lazyTransduce = lazyTransduce;
//# sourceMappingURL=iterables.js.map