"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var propertyNames_1 = require("./propertyNames");
var transducers_1 = require("./transducers");
var util_1 = require("./util");
// Transformers with no parameters, such as the one for count() here, are
// created the first time they are called so they can be tree shaken if unused.
// Tree shaking does not remove top-level object literal constants if they have
// computed keys.
var countTransformer;
function count() {
    var _a;
    if (!countTransformer) {
        countTransformer = (_a = {},
            _a[propertyNames_1.INIT] = function () { return 0; },
            _a[propertyNames_1.RESULT] = function (result) { return result; },
            _a[propertyNames_1.STEP] = function (result) { return result + 1; },
            _a);
    }
    return countTransformer;
}
exports.count = count;
function every(pred) {
    return transducers_1.remove(pred)(isEmpty());
}
exports.every = every;
function find(pred) {
    return transducers_1.filter(pred)(first());
}
exports.find = find;
var firstTransformer;
function first() {
    var _a;
    if (!firstTransformer) {
        firstTransformer = (_a = {},
            _a[propertyNames_1.INIT] = function () { return null; },
            _a[propertyNames_1.RESULT] = function (result) { return result; },
            _a[propertyNames_1.STEP] = function (_, input) { return util_1.reduced(input); },
            _a);
    }
    return firstTransformer;
}
exports.first = first;
var ForEachTransformer = /** @class */ (function () {
    function ForEachTransformer(f) {
        this.f = f;
    }
    ForEachTransformer.prototype[propertyNames_1.INIT] = function () {
        return undefined;
    };
    ForEachTransformer.prototype[propertyNames_1.RESULT] = function () {
        return undefined;
    };
    ForEachTransformer.prototype[propertyNames_1.STEP] = function (_, input) {
        return this.f(input);
    };
    return ForEachTransformer;
}());
function forEach(f) {
    return new ForEachTransformer(f);
}
exports.forEach = forEach;
var isEmptyTransformer;
function isEmpty() {
    var _a;
    if (!isEmptyTransformer) {
        isEmptyTransformer = (_a = {},
            _a[propertyNames_1.INIT] = function () { return true; },
            _a[propertyNames_1.RESULT] = function (result) { return result; },
            _a[propertyNames_1.STEP] = function () { return util_1.reduced(false); },
            _a);
    }
    return isEmptyTransformer;
}
exports.isEmpty = isEmpty;
var JoinToString = /** @class */ (function () {
    function JoinToString(separator) {
        this.separator = separator;
    }
    JoinToString.prototype[propertyNames_1.INIT] = function () {
        return [];
    };
    JoinToString.prototype[propertyNames_1.RESULT] = function (result) {
        return result.join(this.separator);
    };
    JoinToString.prototype[propertyNames_1.STEP] = function (result, input) {
        result.push(input);
        return result;
    };
    return JoinToString;
}());
function joinToString(separator) {
    return new JoinToString(separator);
}
exports.joinToString = joinToString;
var isNotEmptyTransformer;
function some(pred) {
    var _a;
    if (!isNotEmptyTransformer) {
        isNotEmptyTransformer = (_a = {},
            _a[propertyNames_1.INIT] = function () { return false; },
            _a[propertyNames_1.RESULT] = function (result) { return result; },
            _a[propertyNames_1.STEP] = function () { return util_1.reduced(true); },
            _a);
    }
    return transducers_1.filter(pred)(isNotEmptyTransformer);
}
exports.some = some;
var toArrayTransformer;
function toArray() {
    var _a;
    if (!toArrayTransformer) {
        toArrayTransformer = (_a = {},
            _a[propertyNames_1.INIT] = function () { return []; },
            _a[propertyNames_1.RESULT] = function (result) { return result; },
            _a[propertyNames_1.STEP] = function (result, input) {
                result.push(input);
                return result;
            },
            _a);
    }
    return toArrayTransformer;
}
exports.toArray = toArray;
var ToMap = /** @class */ (function () {
    function ToMap(getKey, getValue) {
        this.getKey = getKey;
        this.getValue = getValue;
    }
    ToMap.prototype[propertyNames_1.INIT] = function () {
        return new Map();
    };
    ToMap.prototype[propertyNames_1.RESULT] = function (result) {
        return result;
    };
    ToMap.prototype[propertyNames_1.STEP] = function (result, item) {
        result.set(this.getKey(item), this.getValue(item));
        return result;
    };
    return ToMap;
}());
function toMap(getKey, getValue) {
    return new ToMap(getKey, getValue);
}
exports.toMap = toMap;
var InProgressTransformer = /** @class */ (function () {
    function InProgressTransformer(xf) {
        this.xf = xf;
        this.isReduced = false;
        this.result = xf[propertyNames_1.INIT]();
    }
    InProgressTransformer.prototype.step = function (input) {
        if (!this.isReduced) {
            var newResult = this.xf[propertyNames_1.STEP](this.result, input);
            if (util_1.isReduced(newResult)) {
                this.result = newResult[propertyNames_1.VALUE];
                this.isReduced = true;
            }
            else {
                this.result = newResult;
            }
        }
    };
    InProgressTransformer.prototype.getResult = function () {
        return this.xf[propertyNames_1.RESULT](this.result);
    };
    return InProgressTransformer;
}());
var ToMapGroupBy = /** @class */ (function () {
    function ToMapGroupBy(getKey, xf) {
        this.getKey = getKey;
        this.xf = xf;
    }
    ToMapGroupBy.prototype[propertyNames_1.INIT] = function () {
        return new Map();
    };
    ToMapGroupBy.prototype[propertyNames_1.RESULT] = function (result) {
        var completeResult = new Map();
        var entries = result.entries();
        for (var step = entries.next(); !step.done; step = entries.next()) {
            var _a = step.value, key = _a[0], value = _a[1];
            completeResult.set(key, value.getResult());
        }
        return completeResult;
    };
    ToMapGroupBy.prototype[propertyNames_1.STEP] = function (result, item) {
        var key = this.getKey(item);
        if (!result.has(key)) {
            result.set(key, new InProgressTransformer(this.xf));
        }
        result.get(key).step(item);
        return result;
    };
    return ToMapGroupBy;
}());
function toMapGroupBy(getKey, transformer) {
    if (transformer === void 0) { transformer = toArray(); }
    return new ToMapGroupBy(getKey, transformer);
}
exports.toMapGroupBy = toMapGroupBy;
var ToObject = /** @class */ (function () {
    function ToObject(getKey, getValue) {
        this.getKey = getKey;
        this.getValue = getValue;
    }
    ToObject.prototype[propertyNames_1.INIT] = function () {
        return {};
    };
    ToObject.prototype[propertyNames_1.RESULT] = function (result) {
        return result;
    };
    ToObject.prototype[propertyNames_1.STEP] = function (result, item) {
        result[this.getKey(item)] = this.getValue(item);
        return result;
    };
    return ToObject;
}());
function toObject(getKey, getValue) {
    return new ToObject(getKey, getValue);
}
exports.toObject = toObject;
var ToObjectGroupBy = /** @class */ (function () {
    function ToObjectGroupBy(getKey, xf) {
        this.getKey = getKey;
        this.xf = xf;
    }
    ToObjectGroupBy.prototype[propertyNames_1.INIT] = function () {
        return {};
    };
    ToObjectGroupBy.prototype[propertyNames_1.RESULT] = function (result) {
        var completeResult = {};
        Object.keys(result).forEach(function (key) { return (completeResult[key] = result[key].getResult()); });
        return completeResult;
    };
    ToObjectGroupBy.prototype[propertyNames_1.STEP] = function (result, item) {
        var key = this.getKey(item);
        if (!result[key]) {
            result[key] = new InProgressTransformer(this.xf);
        }
        result[key].step(item);
        return result;
    };
    return ToObjectGroupBy;
}());
function toObjectGroupBy(getKey, transformer) {
    if (transformer === void 0) { transformer = toArray(); }
    return new ToObjectGroupBy(getKey, transformer);
}
exports.toObjectGroupBy = toObjectGroupBy;
var toSetTransformer;
function toSet() {
    var _a;
    if (!toSetTransformer) {
        toSetTransformer = (_a = {},
            _a[propertyNames_1.INIT] = function () { return new Set(); },
            _a[propertyNames_1.RESULT] = function (result) { return result; },
            _a[propertyNames_1.STEP] = function (result, input) {
                result.add(input);
                return result;
            },
            _a);
    }
    return toSetTransformer;
}
exports.toSet = toSet;
var averageTransformer;
function toAverage() {
    var _a;
    if (!averageTransformer) {
        averageTransformer = (_a = {},
            _a[propertyNames_1.INIT] = function () { return [0, 0]; },
            _a[propertyNames_1.RESULT] = function (result) {
                return result[1] === 0 ? null : result[0] / result[1];
            },
            _a[propertyNames_1.STEP] = function (result, input) {
                result[0] += input;
                result[1]++;
                return result;
            },
            _a);
    }
    return averageTransformer;
}
exports.toAverage = toAverage;
var Min = /** @class */ (function () {
    function Min(comparator) {
        this.comparator = comparator;
    }
    Min.prototype[propertyNames_1.INIT] = function () {
        return null;
    };
    Min.prototype[propertyNames_1.RESULT] = function (result) {
        return result;
    };
    Min.prototype[propertyNames_1.STEP] = function (result, input) {
        return result === null || this.comparator(input, result) < 0
            ? input
            : result;
    };
    return Min;
}());
function invertComparator(comparator) {
    return function (a, b) { return -comparator(a, b); };
}
var NATURAL_COMPARATOR = function (a, b) {
    if (a < b) {
        return -1;
    }
    else {
        return a > b ? 1 : 0;
    }
};
function toMax(comparator) {
    if (comparator === void 0) { comparator = NATURAL_COMPARATOR; }
    return new Min(invertComparator(comparator));
}
exports.toMax = toMax;
function toMin(comparator) {
    if (comparator === void 0) { comparator = NATURAL_COMPARATOR; }
    return new Min(comparator);
}
exports.toMin = toMin;
var sumTransformer;
function toSum() {
    var _a;
    if (!sumTransformer) {
        sumTransformer = (_a = {},
            _a[propertyNames_1.INIT] = function () { return 0; },
            _a[propertyNames_1.RESULT] = function (result) { return result; },
            _a[propertyNames_1.STEP] = function (result, input) {
                return result + input;
            },
            _a);
    }
    return sumTransformer;
}
exports.toSum = toSum;
//# sourceMappingURL=reducers.js.map